# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import os
import logging
from typing import Optional

import zmq
from zmq import NOBLOCK, ZMQError, EINVAL, EHOSTUNREACH

from volttron.server.router.servicepeer import ServicePeerNotifier
from volttron.utils.frame_serialization import serialize_frames

__all__ = ["BaseRouter", "OUTGOING", "INCOMING", "UNROUTABLE", "ERROR"]

OUTGOING = 0
INCOMING = 1
UNROUTABLE = 2
ERROR = 3

_log = logging.getLogger(__name__)

# Optimizing by pre-creating frames
_ROUTE_ERRORS = {
    errnum: (
        zmq.Frame(str(errnum).encode("ascii")),
        zmq.Frame(os.strerror(errnum).encode("ascii")),
    )
    for errnum in [zmq.EHOSTUNREACH, zmq.EAGAIN]
}
_INVALID_SUBSYSTEM = (
    zmq.Frame(str(zmq.EPROTONOSUPPORT).encode("ascii")),
    zmq.Frame(os.strerror(zmq.EPROTONOSUPPORT).encode("ascii")),
)


class BaseRouter(object):
    """Abstract base class of VIP router implementation.

    Router implementers should inherit this class and implement the
    setup() method to bind to appropriate addresses, set identities,
    setup authentication, etc, etc. The socket will be created by the
    start() method, which will then call the setup() method.  Once
    started, the socket may be polled for incoming messages and those
    messages are handled/routed by calling the route() method.  During
    routing, the issue() method, which may be implemented, will be
    called to allow for debugging and logging. Custom subsystems may be
    implemented in the handle_subsystem() method. The socket will be
    closed when the stop() method is called.
    """

    _context_class = zmq.Context
    _socket_class = zmq.Socket
    _poller_class = zmq.Poller

    def __init__(
        self,
        context=None,
        default_user_id=None,
        service_notifier=Optional[ServicePeerNotifier],
    ):
        """Initialize the object instance.

        If context is None (the default), the zmq global context will be
        used for socket creation.
        """
        self.context = context or self._context_class.instance()
        self.default_user_id = default_user_id
        self.socket = None
        self._peers = set()
        self._poller = self._poller_class()
        self._ext_sockets = []
        self._socket_id_mapping = {}
        self._service_notifier = service_notifier

    def run(self):
        """Main router loop."""
        self.start()
        try:
            while True:
                self.poll_sockets()
        finally:
            self.stop()

    def start(self):
        """Create the socket and call setup().

        The socket is save in the socket attribute. The setup() method
        is called at the end of the method to perform additional setup.
        """
        self.socket = sock = self._socket_class(self.context, zmq.ROUTER)
        sock.router_mandatory = True
        sock.sndtimeo = 0
        sock.tcp_keepalive = True
        sock.tcp_keepalive_idle = 180
        sock.tcp_keepalive_intvl = 20
        sock.tcp_keepalive_cnt = 6
        self.context.set(zmq.MAX_SOCKETS, 30690)
        sock.set_hwm(6000)
        _log.debug("ROUTER SENDBUF: {0}, {1}".format(sock.getsockopt(zmq.SNDBUF),
                                                     sock.getsockopt(zmq.RCVBUF)))
        self.setup()

    def stop(self, linger=1):
        """Close the socket."""
        self.socket.close(linger)

    def setup(self):
        """Called from start() method to setup the socket.

        Implement this method to bind the socket, set identities and
        options, etc.
        """
        raise NotImplementedError()

    def poll_sockets(self):
        """Called inside run method

        Implement this method to poll for sockets for incoming messages.
        """
        raise NotImplementedError()

    @property
    def poll(self):
        """Returns the underlying socket's poll method."""
        return self.socket.poll

    def handle_subsystem(self, frames, user_id):
        """Handle additional subsystems and provide a response.

        This method does nothing by default and may be implemented by
        subclasses to provide additional subsystems.

        frames is a list of zmq.Frame objects with the following
        elements:

          [SENDER, RECIPIENT, PROTOCOL, USER_ID, MSG_ID, SUBSYSTEM, ...]

        The return value should be None, if the subsystem is unknown, an
        empty list or False (or other False value) if the message was
        handled but does not require/generate a response, or a list of
        containing the following elements:

          [RECIPIENT, SENDER, PROTOCOL, USER_ID, MSG_ID, SUBSYSTEM, ...]

        """
        pass

    def issue(self, topic, frames, extra=None):
        pass

    if zmq.zmq_version_info() >= (4, 1, 0):

        def lookup_user_id(self, sender, recipient, auth_token):
            """Find and return a user identifier.

            Returns the UTF-8 encoded User-Id property from the sender
            frame or None if the authenticator did not set the User-Id
            metadata. May be extended to perform additional lookups.
            """
            # pylint: disable=unused-argument
            # A user id might/should be set by the ZAP authenticator
            try:
                # _log.debug(f"THE TYPE IS:::::::: {type(recipient)}")
                # recipient.get('User-Id').encode('utf-8') returns sender !!!
                return sender
            except ZMQError as exc:
                if exc.errno != EINVAL:
                    raise
            return self.default_user_id

    else:

        def lookup_user_id(self, sender, recipient, auth_token):
            """Find and return a user identifier.

            A no-op by default, this method must be overridden to map
            the sender and auth_token to a user ID. The returned value
            must be a string or None (if the token was not found).
            """
            return self.default_user_id

    def _distribute(self, *parts):
        drop = set()
        empty = ""
        frames = [empty, empty, "VIP1", empty, empty]
        frames.extend(parts)
        # _log.debug(f"_distribute {parts}")
        for peer in self._peers:
            frames[0] = peer
            drop.update(self._send(frames))
        for peer in drop:
            self._drop_peer(peer)
            if self._service_notifier:
                self._service_notifier.peer_dropped(peer)

    def _drop_pubsub_peers(self, peer):
        """Drop peers for pubsub subsystem. To be handled by subclasses"""
        pass

    def _add_pubsub_peers(self, peer):
        """Add peers for pubsub subsystem. To be handled by subclasses"""
        pass

    def _add_peer(self, peer):
        if peer in self._peers:
            return
        self._distribute("peerlist", "add", peer)
        self._peers.add(peer)
        self._add_pubsub_peers(peer)
        if self._service_notifier:
            self._service_notifier.peer_added(peer)

    def _drop_peer(self, peer):
        try:
            self._peers.remove(peer)
        except KeyError:
            return
        self._distribute(b"peerlist", b"drop", peer)
        self._drop_pubsub_peers(peer)

    def route(self, frames):
        """Route one message and return.

        One message is read from the socket and processed. If the
        recipient is the router (empty recipient), the standard hello
        and ping subsystems are handled. Other subsystems are sent to
        handle_subsystem() for processing. Messages destined for other
        entities are routed appropriately.
        """
        socket = self.socket
        issue = self.issue

        issue(INCOMING, frames)
        # _log.debug(f"ROUTER Receiving frames: {frames}")
        if len(frames) < 6:
            # Cannot route if there are insufficient frames, such as
            # might happen with a router probe.
            if len(frames) == 2 and frames[0] and not frames[1]:
                issue(UNROUTABLE, frames, "router probe")
                self._add_peer(frames[0])
            else:
                issue(UNROUTABLE, frames, "too few frames")
            return
        sender, recipient, proto, auth_token, msg_id = frames[:5]
        # _log.debug(f"routing {sender}, {recipient}, {proto}, {auth_token}, {msg_id}")
        if proto != "VIP1":
            # Peer is not talking a protocol we understand
            issue(UNROUTABLE, frames, "bad VIP signature")
            return
        user_id = self.lookup_user_id(sender, recipient, auth_token)
        if user_id is None:
            user_id = ""
        # _log.debug(f"user_id is {user_id}")
        self._add_peer(sender)
        subsystem = frames[5]
        if not recipient:
            # Handle requests directed at the router
            name = subsystem
            if name == "hello":
                frames = [
                    sender,
                    recipient,
                    proto,
                    user_id,
                    msg_id,
                    "hello",
                    "welcome",
                    "1.0",
                    socket.identity,
                    sender,
                ]
            elif name == "ping":
                frames[:7] = [
                    sender,
                    recipient,
                    proto,
                    user_id,
                    msg_id,
                    "ping",
                    "pong",
                ]
            elif name == "peerlist":
                try:
                    op = frames[6]
                except IndexError:
                    op = None
                frames = [sender, recipient, proto, "", msg_id, subsystem]
                if op == "list":
                    frames.append("listing")
                    frames.extend(self._peers)
                else:
                    error = ("unknown" if op else "missing") + " operation"
                    frames.extend(["error", error])
            elif name == "error":
                return
            else:
                response = self.handle_subsystem(frames, user_id)
                if response is None:
                    # Handler does not know of the subsystem
                    errnum, errmsg = error = _INVALID_SUBSYSTEM
                    issue(ERROR, frames, error)
                    frames = [
                        sender,
                        recipient,
                        proto,
                        "",
                        msg_id,
                        "error",
                        errnum,
                        errmsg,
                        "",
                        subsystem,
                    ]
                elif not response:
                    # Subsystem does not require a response
                    return
                else:
                    frames = response
        else:
            # Route all other requests to the recipient
            frames[:4] = [recipient, sender, proto, user_id]
        for peer in self._send(frames):
            self._drop_peer(peer)

    def _send(self, frames):
        issue = self.issue
        socket = self.socket
        drop = []
        recipient, sender = frames[:2]
        # Expecting outgoing frames:
        #   [RECIPIENT, SENDER, PROTO, USER_ID, MSG_ID, SUBSYS, ...]
        try:
            # Try sending the message to its recipient
            # This is a zmq socket so we need to serialize it before sending
            serialized_frames = serialize_frames(frames)
            socket.send_multipart(serialized_frames, flags=NOBLOCK, copy=False)
            issue(OUTGOING, serialized_frames)
        except ZMQError as exc:
            try:
                errnum, errmsg = error = _ROUTE_ERRORS[exc.errno]
            except KeyError:
                error = None
            if error is None:
                raise
            issue(ERROR, frames, error)
            if exc.errno == EHOSTUNREACH:
                drop.append(recipient)
            if exc.errno != EHOSTUNREACH or sender is not frames[0]:
                # Only send errors if the sender and recipient differ
                proto, user_id, msg_id, subsystem = frames[2:6]
                frames = [
                    sender,
                    "",
                    proto,
                    user_id,
                    msg_id,
                    "error",
                    errnum,
                    errmsg,
                    recipient,
                    subsystem,
                ]
                serialized_frames = serialize_frames(frames)
                try:
                    socket.send_multipart(serialized_frames, flags=NOBLOCK, copy=False)
                    issue(OUTGOING, serialized_frames)
                except ZMQError as exc:
                    try:
                        errnum, errmsg = error = _ROUTE_ERRORS[exc.errno]
                    except KeyError:
                        error = None
                    if error is None:
                        raise
                    issue(ERROR, serialized_frames, error)
                    if exc.errno == EHOSTUNREACH:
                        drop.append(sender)
        return drop



# # -*- coding: utf-8 -*- {{{
# # vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
# #
# # Copyright 2020, Battelle Memorial Institute.
# #
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# # http://www.apache.org/licenses/LICENSE-2.0
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.
# #
# # This material was prepared as an account of work sponsored by an agency of
# # the United States Government. Neither the United States Government nor the
# # United States Department of Energy, nor Battelle, nor any of their
# # employees, nor any jurisdiction or organization that has cooperated in the
# # development of these materials, makes any warranty, express or
# # implied, or assumes any legal liability or responsibility for the accuracy,
# # completeness, or usefulness or any information, apparatus, product,
# # software, or process disclosed, or represents that its use would not infringe
# # privately owned rights. Reference herein to any specific commercial product,
# # process, or service by trade name, trademark, manufacturer, or otherwise
# # does not necessarily constitute or imply its endorsement, recommendation, or
# # favoring by the United States Government or any agency thereof, or
# # Battelle Memorial Institute. The views and opinions of authors expressed
# # herein do not necessarily state or reflect those of the
# # United States Government or any agency thereof.
# #
# # PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# # BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# # under Contract DE-AC05-76RL01830
# # }}}
#
# TODO: https://github.com/VOLTTRON/volttron-core/issues/2
# from __future__ import print_function, absolute_import
#
# import logging
# import requests
# import random
# import os
# import grequests
# from datetime import datetime, timedelta
# from zmq import ZMQError
# from volttron.utils import jsonapi
# from gevent.lock import Semaphore
#
# # from volttron.platform.agent import utils
# from volttron.client.vip.agent import Agent, Core
#
# from requests.exceptions import ConnectionError
# from urllib3.exceptions import NewConnectionError
# from urllib.parse import urlparse, urljoin
# from gevent.fileobject import FileObject
# from volttron.utils.persistance import PersistentDict
# from volttron.utils import get_aware_utc_now
#
#
# _log = logging.getLogger(__name__)
#
# __version__ = "0.1"
#
#
# class DiscoveryError(Exception):
#     """Raised when a different volttron central tries to register."""
#
#     pass
#
#
# class KeyDiscoveryAgent(Agent):
#     """
#     Class to get server key, instance name and vip address of external/remote platforms
#     """
#
#     def __init__(
#         self,
#         address,
#         serverkey,
#         identity,
#         external_address_config,
#         setup_mode,
#         bind_web_address,
#         *args,
#         **kwargs
#     ):
#         super(KeyDiscoveryAgent, self).__init__(identity, address, **kwargs)
#         self._external_address_file = external_address_config
#         self._ext_addresses = dict()
#         self._grnlets = dict()
#         self._vip_socket = None
#         self._my_web_address = bind_web_address
#         self.r = random.random()
#         self._setup_mode = setup_mode
#         if self._setup_mode:
#             _log.debug("RUNNING IN MULTI-PLATFORM SETUP MODE")
#
#         self._store_path = os.path.join(
#             os.environ["VOLTTRON_HOME"], "external_platform_discovery.json"
#         )
#         self._ext_addresses_store = dict()
#         self._ext_addresses_store_lock = Semaphore()
#
#     @Core.receiver("onstart")
#     def startup(self, sender, **kwargs):
#         """
#         Try to get platform discovery info of all the remote platforms. If unsuccessful, setup events to try again later
#         :param sender: caller
#         :param kwargs: optional arguments
#         :return:
#         """
#         self._vip_socket = self.core.socket
#
#         # If in setup mode, read the external_addresses.json to get web addresses to set up authorized connection with
#         # external platforms.
#         if self._setup_mode:
#             if self._my_web_address is None:
#                 _log.error("Web bind address is needed in multiplatform setup mode")
#                 return
#             with self._ext_addresses_store_lock:
#                 try:
#                     self._ext_addresses_store = PersistentDict(
#                         filename=self._store_path, flag="c", format="json"
#                     )
#                 except ValueError as exc:
#                     _log.error("Error in json format: {0}".format(exc))
#                 # Delete the existing store.
#                 if self._ext_addresses_store:
#                     self._ext_addresses_store.clear()
#                     self._ext_addresses_store.async_sync()
#             web_addresses = dict()
#             # Read External web addresses file
#             try:
#                 web_addresses = self._read_platform_address_file()
#                 try:
#                     web_addresses.remove(self._my_web_address)
#                 except ValueError:
#                     _log.debug(
#                         "My web address is not in the external bind web adress list"
#                     )
#
#                 op = "web-addresses"
#                 self._send_to_router(op, web_addresses)
#             except IOError as exc:
#                 _log.error("Error in reading file: {}".format(exc))
#                 return
#             sec = random.random() * self.r + 10
#             delay = get_aware_utc_now() + timedelta(seconds=sec)
#             grnlt = self.core.schedule(delay, self._key_collection, web_addresses)
#         else:
#             # Use the existing store for platform discovery information
#             with self._ext_addresses_store_lock:
#                 try:
#                     self._ext_addresses_store = PersistentDict(
#                         filename=self._store_path, flag="c", format="json"
#                     )
#                 except ValueError as exc:
#                     _log.error("Error in json file format: {0}".format(exc))
#                 for name, discovery_info in self._ext_addresses_store.items():
#                     op = "normalmode_platform_connection"
#                     self._send_to_router(op, discovery_info)
#
#     def _key_collection(self, web_addresses):
#         """
#         Collect platform discovery information (server key, instance name, vip-address) for all platforms.
#         :param web_addresses: List of web addresses to get discovery info
#         :return:
#         """
#         for web_address in web_addresses:
#             if web_address not in self._my_web_address:
#                 self._collect_key(web_address)
#
#     def _collect_key(self, web_address):
#         """
#         Try to get (server key, instance name, vip-address) of remote instance and send it to RoutingService
#         to connect to the remote instance. If unsuccessful, try again later.
#         :param name: instance name
#         :param web_address: web address of remote instance
#         :return:
#         """
#         platform_info = dict()
#
#         try:
#             platform_info = self._get_platform_discovery(web_address)
#             with self._ext_addresses_store_lock:
#                 _log.debug("Platform discovery info: {}".format(platform_info))
#                 name = platform_info["instance-name"]
#                 self._ext_addresses_store[name] = platform_info
#                 self._ext_addresses_store.async_sync()
#         except KeyError as exc:
#             _log.error("Discovery info does not contain instance name {}".format(exc))
#         except DiscoveryError:
#             # If discovery error, try again later
#             sec = random.random() * self.r + 30
#             delay = utils.get_aware_utc_now() + timedelta(seconds=sec)
#             grnlet = self.core.schedule(delay, self._collect_key, web_address)
#         except ConnectionError as e:
#             _log.error("HTTP connection error {}".format(e))
#
#         # If platform discovery is successful, send the info to RoutingService
#         # to establish connection with remote platform.
#         if platform_info:
#             op = "setupmode_platform_connection"
#             connection_settings = dict(platform_info)
#             connection_settings["web-address"] = web_address
#             self._send_to_router(op, connection_settings)
#
#     def _send_to_router(self, op, platform_info):
#         """
#         Send the platform discovery stats to the router to establish new connection
#         :param platform_info: platform discovery stats
#         :return:
#         """
#         address = jsonapi.dumps(platform_info)
#
#         frames = [op, address]
#         try:
#             self._vip_socket.send_vip("", "routing_table", frames, copy=False)
#         except ZMQError as ex:
#             # Try sending later
#             _log.error(
#                 "ZMQ error while sending external platform info to router: {}".format(
#                     ex
#                 )
#             )
#
#     def _read_platform_address_file(self):
#         """
#         Read the external addresses file
#         :return:
#         """
#
#         try:
#             with open(self._external_address_file) as fil:
#                 # Use gevent FileObject to avoid blocking the thread
#                 data = FileObject(fil, close=False).read()
#                 web_addresses = jsonapi.loads(data) if data else {}
#                 return web_addresses
#         except IOError as e:
#             _log.error("Error opening file {}".format(self._external_address_file))
#             raise
#         except Exception:
#             _log.exception("error loading %s", self._external_address_file)
#             raise
#
#     def _get_platform_discovery(self, web_address):
#         """
#         Use http discovery call to get serverkey, instance name and vip-address of remote instance
#         :param web_address: web address of remote instance
#         :return:
#         """
#
#         r = {}
#         try:
#             parsed = urlparse(web_address)
#
#             assert parsed.scheme
#             assert not parsed.path
#
#             real_url = urljoin(web_address, "/discovery/")
#             req = grequests.get(real_url)
#             responses = grequests.map([req])
#             responses[0].raise_for_status()
#             r = responses[0].json()
#             return r
#         except requests.exceptions.HTTPError:
#             raise DiscoveryError("Invalid discovery response from {}".format(real_url))
#         except requests.exceptions.Timeout:
#             raise DiscoveryError("Timeout error from {}".format(real_url))
#         except AttributeError as e:
#             raise DiscoveryError("Invalid web_address passed {}".format(web_address))
#         except (ConnectionError, NewConnectionError) as e:
#             raise DiscoveryError("Connection to {} not available".format(real_url))
#         except Exception as e:
#             raise DiscoveryError("Unknown Exception: {}".format(e))



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""VOLTTRON platform™ topic templates.

Templates of standard topics.  Fields in the templates are replaced by
calling the template with the field value included in the keyword
arguments.  Fields are replaced from left to right as long as a
replacement can be made.  Once a field is reached which cannot be
replaced, everything in the replaced portion up to the last double slash
is returned.  Fields cannot be skipped, but may be included
unsubstituted by using None for the field value.  Below are some
examples to demonstrate.

    >>> T = _('root/{top}//{middle}//{bottom}')
    >>> T()
    Topic(u'root')
    >>> T(top='first')
    Topic(u'root/first')
    >>> T(top='first', middle='second')
    Topic(u'root/first/second')
    >>> T(top='first', middle='second', bottom='third')
    Topic(u'root/first/second/third')
    >>> unicode(T(top='first', middle='second', bottom='third'))
    u'root/first/second/third'
    >>> T(top='first', bottom='third')
    ValueError: unused keyword argument: bottom
    >>> T(top='first', middle=None, bottom='third')
    Topic(u'root/first/{middle}/third')
    >>> T(top='first', middle=None, bottom='third')(middle='.')
    Topic(u'root/first/third')
    >>> T(top='first', middle=None, bottom='third')(middle='..')
    Topic(u'root/third')
    >>> T._(top='first', middle=None, bottom='third')
    Topic(u'root/first//{middle}//third')
"""

import os
import platform

from .utils import Topic as _

__author__ = "Brandon Carpenter <brandon.carpenter@pnnl.gov>"
__copyright__ = "Copyright (c) 2016, Battelle Memorial Institute"
__license__ = "Apache 2.0"

ALERTS_BASE = _("alerts")
ALERTS = _("alerts/{agent_class}/{agent_identity}")    # /{agent_class}/{publickey}/{alert_key}')

HEARTBEAT = _("heartbeats")
PLATFORM_BASE = _("platform")
PLATFORM_SEND_EMAIL = _("platform/send_email")
PLATFORM = _("platform/{subtopic}")
PLATFORM_SHUTDOWN = PLATFORM(subtopic="shutdown")
PLATFORM_VCP_DEVICES = _("platforms/{platform_uuid}/devices/{topic}")

RECORD_BASE = _("record")
RECORD = _("record/{subtopic}")

MARKET_BASE = _("market/{subtopic}")
MARKET_RESERVE = _(MARKET_BASE.replace("{subtopic}", "reserve"))
MARKET_BID = _(MARKET_BASE.replace("{subtopic}", "bid"))
MARKET_CLEAR = _(MARKET_BASE.replace("{subtopic}", "cleared_price"))
MARKET_AGGREGATE = _(MARKET_BASE.replace("{subtopic}", "aggregate"))
MARKET_ERROR = _(MARKET_BASE.replace("{subtopic}", "error"))
MARKET_RECORD = _(RECORD.replace("{subtopic}", MARKET_CLEAR))

AGENT_SHUTDOWN = _("agent/{agent}/shutdown")
AGENT_PING = _("agent/ping/{}/{}/{{cookie}}".format(platform.uname()[1], os.getpid()))

LOGGER_BASE = _("datalogger")
LOGGER = _("datalogger/{subtopic}")
LOGGER_LOG = LOGGER(subtopic="log")
LOGGER_STATUS = LOGGER(subtopic="status")

DRIVER_TOPIC_BASE = "devices"
DRIVER_TOPIC_ALL = "all"
DEVICES_PATH = _("{base}//{node}//{campus}//{building}//{unit}//{path!S}//{point}")
_DEVICES_VALUE = _(DEVICES_PATH.replace("{base}", DRIVER_TOPIC_BASE))
DEVICES_VALUE = _(_DEVICES_VALUE.replace("{node}/", ""))

# For use with RPC calls that require a device path. A plain device path with no prefix.
# Should be used when working with devices via the actuator agent RPC calls:
# get_point, set_point, revert_point, revert_device, and request_new_schedule.
RPC_DEVICE_PATH = _(DEVICES_PATH.replace("{base}//{node}//", ""))

ANALYSIS_PATH = _("{base}//{analysis_name}//{campus}//{building}//{unit}//{point}")
ANALYSIS_TOPIC_BASE = "analysis"
ANALYSIS_VALUE = _(ANALYSIS_PATH.replace("{base}", ANALYSIS_TOPIC_BASE))

ACTUATOR_GET = _(_DEVICES_VALUE.replace("{node}", "actuators/get"))
ACTUATOR_SET = _(_DEVICES_VALUE.replace("{node}", "actuators/set"))
ACTUATOR_REVERT_POINT = _(_DEVICES_VALUE.replace("{node}", "actuators/revert/point"))
ACTUATOR_REVERT_DEVICE = _(_DEVICES_VALUE.replace("{node}", "actuators/revert/device"))

_ACTUATOR_SCHEDULE = _(("{base}/actuators/schedule/{op}").replace("{base}", DRIVER_TOPIC_BASE))
ACTUATOR_SCHEDULE_REQUEST = _(_ACTUATOR_SCHEDULE.replace("{op}", "request"))
ACTUATOR_SCHEDULE_RESULT = _(_ACTUATOR_SCHEDULE.replace("{op}", "result"))
ACTUATOR_SCHEDULE_ANNOUNCE_RAW = _(_ACTUATOR_SCHEDULE.replace("{op}", "announce/{device}"))

# This is a convenience topic for agent listening for announcements
# and want to use the {campus}//{building}//{unit} style replacement
ACTUATOR_SCHEDULE_ANNOUNCE = _(
    ACTUATOR_SCHEDULE_ANNOUNCE_RAW.replace("{device}", "{campus}//{building}//{unit}"))

# Added by CHA to be used as the root of all actuators for working within
# base_historian.py.
ACTUATOR_BASE = _("actuators")
ACTUATOR = _(_DEVICES_VALUE.replace("{node}", ACTUATOR_BASE))
ACTUATOR_ERROR = _(_DEVICES_VALUE.replace("{node}", "actuators/error"))
ACTUATOR_VALUE = _(_DEVICES_VALUE.replace("{node}", "actuators/value"))

# Ragardless of the interface used (RPC vs pubsub) when an agent
# attempts to set a point it is announced on this topic.
# This is intended to inable a historian to capture all attempted writes.
ACTUATOR_WRITE = _(_DEVICES_VALUE.replace("{node}", "actuators/write"))
ACTUATOR_REVERTED_POINT = _(_DEVICES_VALUE.replace("{node}", "actuators/reverted/point"))
ACTUATOR_REVERTED_DEVICE = _(_DEVICES_VALUE.replace("{node}", "actuators/reverted/device"))

BASE_ARCHIVER_REQUEST = _("archiver/request")
BASE_ARCHIVER_FULL_REQUEST = _("archiver/full/request")
BASE_ARCHIVER_RESPONSE = _("archiver/response")

_ARCHIVER = _("{base}/{campus}//{building}//{unit}//{point}")
_ARCHIVER_UNIT = _("{base}/{campus}//{building}//{unit}")
ARCHIVER_REQUEST = _(_ARCHIVER.replace("{base}", BASE_ARCHIVER_REQUEST))
ARCHIVER_RESPONSE = _(_ARCHIVER.replace("{base}", BASE_ARCHIVER_RESPONSE))

ARCHIVER_FULL_UNIT_REQUEST = _(_ARCHIVER_UNIT.replace("{base}", BASE_ARCHIVER_FULL_REQUEST))

OPENADR_STATUS = _("openadr/status")
OPENADR_EVENT = _("openadr/event")

_SUBSCRIPTIONS = _("subscriptions/{op}/{{topic}}")
SUBSCRIPTIONS_LIST = _(_SUBSCRIPTIONS.format(op="list"))
SUBSCRIPTIONS_ADD = _(_SUBSCRIPTIONS.format(op="add"))
SUBSCRIPTIONS_REMOVE = _(_SUBSCRIPTIONS.format(op="remove"))

_BUILDING = _("building/{op}/{{campus}}//{{building}}//{{topic}}")
BUILDING_SEND = _(_BUILDING.format(op="send"))
BUILDING_RECV = _(_BUILDING.format(op="recv"))
BUILDING_ERROR = _(_BUILDING.format(op="error"))

CONFIG_TOPIC_BASE = "config"
CONFIG_PATH = _("{base}//{action}//{category}//{name}")
_CONFIG_VALUE = _(CONFIG_PATH.replace("{base}", CONFIG_TOPIC_BASE))
CONFIG_ADD = _(_CONFIG_VALUE.replace("{action}", "add"))
CONFIG_REMOVE = _(_CONFIG_VALUE.replace("{action}", "remove"))
CONFIG_UPDATE = _(_CONFIG_VALUE.replace("{action}", "update"))

DRIVER_CONFIG_ADD = _(_CONFIG_VALUE.replace("{category}", "driver"))
DRIVER_CONFIG_REMOVE = _(_CONFIG_VALUE.replace("{category}", "driver"))
DRIVER_CONFIG_UPDATE = _(_CONFIG_VALUE.replace("{category}", "driver"))

WEATHER_BASE = "weather"
WEATHER_REQUEST = "weather/request"

WEATHER2_BASE = "weather2"
WEATHER2_REQUEST = "weather2/request"
WEATHER2_POLL = "weather2/polling"
WEATHER2_RESPONSE = "weather2/response"
WEATHER2_ERROR = "weather2/error"

BACNET_INFO_BASE = _("bacnet/{indication}")
BACNET_I_AM = _(BACNET_INFO_BASE.replace("{indication}", "i_am"))



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import weakref

from volttron.client.vip.agent.subsystems.base import SubsystemBase
from volttron.client.messaging.headers import TIMESTAMP
from volttron.utils.time import get_aware_utc_now, format_timestamp
from volttron.utils.scheduling import periodic
from volttron.client.vip.agent.errors import Unreachable
"""The heartbeat subsystem adds an optional periodic publish to all agents.
Heartbeats can be started with agents and toggled on and off at runtime.
"""

__docformat__ = "reStructuredText"
__version__ = "1.0"


class Heartbeat(SubsystemBase):

    def __init__(self, owner, core, rpc, pubsub, heartbeat_autostart, heartbeat_period):
        self.owner = owner
        self.core = weakref.ref(core)
        self.pubsub = weakref.ref(pubsub)

        self.autostart = heartbeat_autostart
        self.period = heartbeat_period
        self.enabled = False
        self.connect_error = False

        def onsetup(sender, **kwargs):
            rpc.export(self.start, "heartbeat.start")
            rpc.export(self.start_with_period, "heartbeat.start_with_period")
            rpc.export(self.stop, "heartbeat.stop")
            rpc.export(self.restart, "heartbeat.restart")
            rpc.export(self.set_period, "heartbeat.set_period")

        def onstart(sender, **kwargs):
            if self.autostart:
                self.start()

        core.onsetup.connect(onsetup, self)
        core.onstart.connect(onstart, self)
        core.onconnected.connect(self.reconnect)

    def start(self):
        """RPC method

        Starts an agent's heartbeat.
        """
        if not self.enabled:
            self.scheduled = self.core().schedule(periodic(self.period), self.publish)
            self.enabled = True

    def start_with_period(self, period):
        """RPC method

        Set period and start heartbeat.

        :param period: Time in seconds between publishes.
        """
        self.set_period(period)
        self.start()

    def reconnect(self, sender, **kwargs):
        if self.connect_error:
            self.restart()
            self.connect_error = False

    def stop(self):
        """RPC method

        Stop an agent's heartbeat.
        """
        if self.enabled:
            # Trap the fact that scheduled may not have been
            # set yet if the start hasn't been called.
            try:
                self.scheduled.cancel()
            except AttributeError:
                pass
            self.enabled = False

    def restart(self):
        """RPC method

        Restart the heartbeat with the current period.  The heartbeat will
        be immediately sending the heartbeat to the message bus.
        """
        self.stop()
        self.start()

    def set_period(self, period):
        """RPC method

        Set heartbeat period.

        :param period: Time in seconds between publishes.
        """
        if self.enabled:
            self.stop()
            self.period = period
            self.start()
        else:
            self.period = period

    def publish(self):
        topic = "heartbeat/" + self.core().identity
        headers = {TIMESTAMP: format_timestamp(get_aware_utc_now())}
        message = self.owner.vip.health.get_status_value()
        try:
            self.pubsub().publish("pubsub", topic, headers, message)
        except Unreachable as exc:
            self.connect_error = True
            self.stop()



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import functools
from types import MethodType

import gevent

__all__ = ["annotate", "annotations", "dualmethod", "spawn"]


def annotate(obj, kind, name, value):
    # pylint: disable=protected-access
    try:
        annotations = obj._annotations
    except AttributeError:
        obj._annotations = annotations = {}
    try:
        items = annotations[name]
    except KeyError:
        annotations[name] = items = kind()
    assert isinstance(items, kind)
    try:
        add = items.add
    except AttributeError:
        try:
            add = items.append
        except AttributeError:
            try:
                add = items.update
            except AttributeError:
                items += value
                return
    add(value)


def annotations(obj, kind, name):
    # pylint: disable=protected-access
    try:
        annotations = obj._annotations
    except AttributeError:
        annotations = {}
    try:
        items = annotations[name]
    except KeyError:
        items = kind()
    assert isinstance(items, kind)
    return items


def spawn(method):
    """Run a decorated method in its own greenlet, which is returned."""

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        return gevent.spawn(method, *args, **kwargs)

    return wrapper


class dualmethod(object):
    """Descriptor to allow class and instance methods of the same name.

    This class implements a descriptor that works similar to the
    classmethod() built-ins and can be used as a decorator, like the
    property() built-in. Instead of a method being only a class or
    instance method, two methods can share the same name and be accessed
    as an instance method or a class method based on the context.

    Example:

    >>> class Foo(object):
    ...     @dualmethod
    ...     def bar(self):
    ...         print('instance method for', self)
    ...     @bar.classmethod
    ...     def bar(cls):
    ...         print('class method for', cls)
    ...
    >>> Foo.bar()
    class method for <class '__main__.Foo'>
    >>> Foo().bar()
    instance method for <__main__.Foo object at 0x7fcd744f6610>
    >>>
    """

    def __init__(self, finstance=None, fclass=None, doc=None):
        """Instantiate the descriptor with the given parameters.

        If finstance is set, it must be a method implementing instance
        access. If fclass is set, it must be a method implementing class
        access similar to a classmethod. If doc is set, it will be used
        for the __doc__ attribute.  Otherwise, the __doc__ attribute
        from the instance or class method will be used, in that order.
        """
        self.finstance = finstance
        self.fclass = fclass
        if doc is not None:
            self.__doc__ = doc
        elif finstance is not None:
            self.__doc__ = finstance.__doc__
        elif fclass is not None:
            self.__doc__ = fclass.__doc__

    def __get__(self, instance, owner):
        """Descriptor getter method.

        See Python descriptor documentation."""
        if instance is None:
            if self.fclass is None:
                if self.finstance is None:
                    raise AttributeError("no instance or class method is set")
                return MethodType(self.finstance, instance)
            return MethodType(self.fclass, owner)
        if self.finstance is None:
            if self.fclass is None:
                raise AttributeError("no instance or class method is set")
            return MethodType(self.fclass, owner)
        return MethodType(self.finstance, instance)

    def instancemethod(self, finstance):
        """Descriptor to set the instance method."""
        self.finstance = finstance
        return self

    def classmethod(self, fclass):
        """Descriptor to set the class method."""
        self.fclass = fclass
        return self



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import weakref

__all__ = ["Signal"]


class Signal(object):

    def __init__(self):
        self._receivers = weakref.WeakValueDictionary()

    def connect(self, receiver, owner=None):
        self._receivers[receiver] = receiver if owner is None else owner

    def disconnect(self, receiver):
        try:
            self._receivers.pop(receiver)
            return True
        except KeyError:
            return False

    def send(self, sender, **kwargs):
        return [receiver(sender, **kwargs) for receiver in self._receivers]

    def sendby(self, executor, sender, **kwargs):
        return [executor(receiver, sender, **kwargs) for receiver in self._receivers]

    def receiver(self, func):
        self.connect(func)
        return func

    def __bool__(self):
        return bool(self._receivers)



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
""" Core package."""
from typing import List

from gevent import monkey

# At this point these are the only things that need to be patched
# and the server and client are working harmoniously with this.
patches = [
    ('ssl', monkey.patch_ssl),
    ('socket', monkey.patch_socket),
    ('os', monkey.patch_os),
]

# patch modules if necessary.  Only if the module hasn't been patched before.
# this could happen if the server code uses the client (which it does).
for module, fn in patches:
    if not monkey.is_module_patched(module):
        fn()

import logging

from urllib.parse import urlparse

from volttron.client.vip.agent import Agent, build_agent
from volttron.client.vip.agent.core import Core
from volttron.client.vip.agent.subsystems.rpc import RPC
from volttron.client.vip.agent.subsystems.channel import Channel
from volttron.client.vip.agent.subsystems.hello import Hello
from volttron.client.vip.agent.subsystems.peerlist import PeerList
from volttron.client.vip.agent.subsystems.ping import Ping
from volttron.client.vip.agent.subsystems.pubsub import PubSub
from volttron.client.vip.agent.subsystems.rpc import RPC
from volttron.client.vip.agent.subsystems.heartbeat import Heartbeat
from volttron.client.vip.agent.subsystems.health import Health
from volttron.client.vip.agent.subsystems.configstore import ConfigStore
from volttron.client.vip.agent.subsystems.auth import Auth
from volttron.client.vip.agent.subsystems.query import Query

__all__: List[str] = [
    "Agent", "Core", "RPC", "Channel", "Hello", "PeerList", "Ping", "PubSub", "Heartbeat",
    "Health", "ConfigStore", "Auth", "Query"
]

_log = logging.getLogger(__name__)


def build_vip_address_string(vip_root, serverkey, publickey, secretkey):
    """Build a full vip address string based upon the passed arguments

    All arguments are required to be non-None in order for the string to be
    created successfully.

    :raises ValueError if one of the parameters is None.
    """
    _log.debug("root: {}, serverkey: {}, publickey: {}, secretkey: {}".format(
        vip_root, serverkey, publickey, secretkey))
    parsed = urlparse(vip_root)
    if parsed.scheme == "tcp":
        if not (serverkey and publickey and secretkey and vip_root):
            raise ValueError("All parameters must be entered.")

        root = "{}?serverkey={}&publickey={}&secretkey={}".format(vip_root, serverkey, publickey,
                                                                  secretkey)

    elif parsed.scheme == "ipc":
        root = vip_root
    else:
        raise ValueError("Invalid vip root specified!")

    return root


# def update_volttron_script_path(path: str) -> str:
#     """
#     Assumes that path's current working directory is in the root directory of the volttron codebase.

#     Prepend 'VOLTTRON_ROOT' to internal volttron script if 'VOLTTRON_ROOT' is set and return new path;
#     otherwise, return original path
#     :param path: relative path to the internal volttron script
#     :return: updated path to volttron script
#     """
#     if os.environ["VOLTTRON_ROOT"]:
#         args = path.split("/")
#         path = f"{os.path.join(os.environ['VOLTTRON_ROOT'], *args)}"
#     _log.debug(f"Path to script: {path}")
#     return path
